#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <OpenSim/OpenSim.h>
#include <InverseDynamicsSolver.h>
//#include <omp.h>

using namespace OpenSim;
using namespace SimTK;
using namespace std;
//#define NUMTHREADS 14

//______________________________________________________________________________


static Model *osimModel;
static State *osimState;
static InverseDynamicsSolver *idSolver;
static bool modelIsLoaded = false;

void ClearMemory(void)
{
	delete osimModel;
	delete idSolver;
	mexPrintf("Cleared memory from opensimDeriv mex file.\n");
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{

	if (nrhs == 1) // Load model if only one right hand side argument
	{
		if (modelIsLoaded == true)
		{
			delete osimModel;
			delete idSolver;
		}
		// extract filenames
        string model_name = mxArrayToString(prhs[0]);
		osimModel = new Model(model_name); // name of Opensim model to load
		// Initialize state
		osimState = &osimModel->initSystem();
		modelIsLoaded = true;
		// Create an inverse dynamics solver
		idSolver = new InverseDynamicsSolver(*osimModel);
		mexAtExit(ClearMemory); // exit function to clear memory allocated in heap
	}
	else if(nrhs == 4) // time, states, controls, options vector
	{
		if (modelIsLoaded == false)
		{
			mexErrMsgTxt("!!!No OpenSim model has been loaded!!!\n");
		}

		const int numPts = mxGetM(prhs[0]); // get number of rows of time vector
		const int numStates = mxGetN(prhs[1]); // get number of states (columns in state matrix)
		const int numControls = mxGetN(prhs[2]); // get number of controls (columns in controls matrix)

		double *time = mxGetPr(prhs[0]); // time vector
		double *x = mxGetPr(prhs[1]); // states matrix
		double *u = mxGetPr(prhs[2]); // controls matrix
		double *options = mxGetPr(prhs[3]); // options vector

		// Declaring variables required for inverse dynamics solver
		int numQs = osimState->getNQ();
		plhs[1] = mxCreateDoubleMatrix(numPts,numQs,mxREAL);
		double *idLoads = mxGetPr(plhs[1]);
		Vector idLoadsVec;
		Vector UDotVec;

		// Declaring variables required for marker positions
		MarkerSet& refMarkerSet = osimModel->updMarkerSet();
		Array<std::string> tmpArrMarkerNames;
		refMarkerSet.getNames(tmpArrMarkerNames); // number of markers
		int numMarkers = tmpArrMarkerNames.getSize();
		int dims[3];
		dims[0] = numPts;
		dims[1] = numMarkers;
		dims[2] = 3; // three coordinates in space per marker
		plhs[2] = mxCreateNumericArray(3, dims, mxDOUBLE_CLASS, mxREAL);
		double* MarkerGlobalPos = mxGetPr(plhs[2]);

		// Declaring variables required for force reporting
		const ForceSet& forces = osimModel->getForceSet(); // This does not contain gravity
        int nf = forces.getSize();
		Storage forceStore;

		if (options[0] == 1) // if compute x_dot set to true
		{
			double *x_row = (double*) mxCalloc(numStates, sizeof(double));
			double *u_row = (double*) mxCalloc(numControls, sizeof(double));

			plhs[0] = mxCreateDoubleMatrix(numPts,numStates,mxREAL);
			double *x_dot = mxGetPr(plhs[0]);

			// #pragma omp parallel for num_threads(NUMTHREADS) // convert loop to parrallel using OpenMP
			// start looping for all collocation points
			for (int i=0; i<numPts; ++i)
			{
				// get a row vector from each of the matrices
				for (int j=0;j<numStates;j++)
				{
					x_row[j] = x[i+numPts*j]; // irow+nrow*icol (column-wise indexing)
				}
				for (int j=0;j<numControls;j++)
				{
					u_row[j] = u[i+numPts*j]; // irow+nrow*icol (column-wise indexing)
				}
				// Set time
				osimState->setTime(time[i]);

				// Set states
				const Vector y(numStates, x_row);
				osimState->updY() = y;

				// Update state velocity calculations
				osimModel->assemble(*osimState);
                // osimModel->calcMassCenterVelocity(*osimState);

				// Define activation control vectors for each actuator
				const Vector newControls(numControls, u_row);

				// Get a writeable reference to model controls and set to new control values
				osimModel->updControls(*osimState) = newControls;
			
				// Done modifying model controls so mark as valid
				osimModel->markControlsAsValid(*osimState);

				// Perform the derivative calculations at this state
				osimModel->computeStateVariableDerivatives(*osimState);

				// Set output variable to new state
				const Vector& ydot = osimState->getYDot();
				for (int j=0; j<numStates; ++j)
				{
					x_dot[i + numPts*j] = ydot[j];
				}

				if (options[1] == 1) // if get residual loads set to true
				{
					// get Udot (Different from states)
					UDotVec = osimState->getUDot();

					// calling InverseDynamicsSolver
					idLoadsVec = idSolver->solve(*osimState,UDotVec); 

					// assign residual loads for each time to output matrix
					for (int j=0;j<numQs;j++)
					{
						idLoads[i + numPts*j] = idLoadsVec[j]; 
					}
				}
				if (options[2] == 1) // if get marker global positions set to true
				{
				for(int j=0; j<numMarkers; j++)
					{
						Marker& tempRefMarker = refMarkerSet.get(tmpArrMarkerNames[j]);
						OpenSim::Body& tempRefMarkerParentBody = tempRefMarker.getBody();
						Vec3 tempMarkerLocalPos = tempRefMarker.getOffset();
						Vec3 tempMarkerGlobalPos;
						osimModel->getSimbodyEngine().getPosition(*osimState, tempRefMarkerParentBody, tempMarkerLocalPos, tempMarkerGlobalPos);
						for (int k=0;k<3;k++)
						{
							MarkerGlobalPos[i + numPts*j+numPts*numMarkers*k] = tempMarkerGlobalPos(k);
						}
					} 
				}
				if (options[3] == 1) // if get force values set to true
				{
					StateVector nextRow = StateVector(osimState->getTime());
					for(int j=0;j<nf;j++)
					{
						OpenSim::Force& nextForce = (OpenSim::Force&)forces[j];
						if (nextForce.isDisabled(*osimState)) continue;
						Array<double> values = nextForce.getRecordValues(*osimState);
						nextRow.getData().append(values);
					}
					forceStore.append(nextRow); 
				}
			}
			if (options[3] == 1) // if get force values set to true
			{
				int numMeasures = forceStore.getSmallestNumberOfStates();
				plhs[3] = mxCreateDoubleMatrix(numPts,numMeasures,mxREAL);
				double *forceMeasures = mxGetPr(plhs[3]);
				double val;
				for (int i=0;i<numPts;i++)
				{
					for (int j=0;j<numMeasures;j++)
					{
						forceStore.getData(i,j,val);
						forceMeasures[i + numPts*j] = val;
					}
				}
			}
			mxFree(x_row); // release memory for vectors of states for each time
			mxFree(u_row);
		}
	}
	else
	{
		mexErrMsgTxt("!!!Wrong inputs or outputs!!!\n");
	}

}


