#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <OpenSim/OpenSim.h>
#include <omp.h>

using namespace OpenSim;
using namespace SimTK;
using namespace std;
#define NTHREADS 14

//______________________________________________________________________________


static Model *osimModel[NTHREADS];
static State *osimState[NTHREADS];
static bool modelIsLoaded = false;

void ClearMemory(void)
{
    for (int i=0; i<NTHREADS; ++i)
                delete osimModel[i];
    modelIsLoaded = false;
    mexPrintf("Cleared memory from opensimGetBodyPosVel_OMP mex file.\n");
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	mexAtExit(ClearMemory); // exit function to clear memory allocated in heap
    
    if (nrhs == 1) // Load model if only one right hand side argument
    {
        if (modelIsLoaded == true)
            ClearMemory();
        // extract filenames
        string model_name = mxArrayToString(prhs[0]);
        for (int i=0; i<NTHREADS; ++i)
        {
            osimModel[i] = new Model(model_name); // name of Opensim model to load
            // Initialize state
            osimState[i] = &osimModel[i]->initSystem();
        }
        modelIsLoaded = true;
    }
    else if(nrhs == 4) // time, states, bodies indices
    {
        if (modelIsLoaded == false)
        {
            mexErrMsgTxt("!!!No OpenSim model has been loaded!!!\n");
        }
        
        const int numPts = mxGetM(prhs[0]); // get number of rows of time vector
        const int numStates = mxGetN(prhs[1]); // get number of states (columns in state matrix)
        const int numBodies = mxGetM(prhs[3]); // get number of bodies
        
        double *time = mxGetPr(prhs[0]); // time vector
        double *x = mxGetPr(prhs[1]); // states matrix
		double *pointOnBody = mxGetPr(prhs[2]); // location of a point on each body
        double *bodies = mxGetPr(prhs[3]); // body number index

		if (numBodies != mxGetM(prhs[2]))
		{
			mexErrMsgTxt("Error: Must have one point per body index given\n");
		}
        
        OpenSim::BodySet *refBodySet[NTHREADS];
        // define variables used in determining spring locations and velocities
        for (int i=0; i<NTHREADS; ++i)
        {
            refBodySet[i] = &osimModel[i]->updBodySet();
        }
        
        // create output matrices
        plhs[0] = mxCreateDoubleMatrix(numPts,numBodies*3,mxREAL);
        plhs[1] = mxCreateDoubleMatrix(numPts,numBodies*3,mxREAL);
		plhs[2] = mxCreateDoubleMatrix(numPts,numBodies*3,mxREAL);
        plhs[3] = mxCreateDoubleMatrix(numPts,numBodies*3,mxREAL);
        double *bodyPos = mxGetPr(plhs[0]);
        double *bodyVel = mxGetPr(plhs[1]);
		double *bodyAng = mxGetPr(plhs[2]);
		double *bodyAngVel = mxGetPr(plhs[3]);
        
        // start looping for all collocation points
        #pragma omp parallel for num_threads(NTHREADS)
        for (int i=0; i<numPts; ++i)
        {
            int thread_id = omp_get_thread_num();

            // change based on number of threads being used
            //double *x_row = new double [numStates];
            double *x_row = (double*) calloc(numStates, sizeof(double));

//             Model *osimModelThread = osimModel[thread_id];
//             State *osimStateThread = osimState[thread_id];
            
            // get a row vector from each of the matrices
            for (int j=0;j<numStates;j++)
            {
                x_row[j] = x[i+numPts*j]; // irow+nrow*icol (column-wise indexing)
            }

            // Set time
            osimState[thread_id]->setTime(time[i]);

            // Set states
            const Vector y(numStates, x_row);
            osimState[thread_id]->updY() = y;

            // Update state velocity calculations
//             osimModel->assemble(*osimState);
            osimModel[thread_id]->calcMassCenterVelocity(*osimState[thread_id]);
            
            // get spring positions and velocities
            for(int j=0; j<numBodies; j++)
            {
                OpenSim::Body& tempRefParentBody = refBodySet[thread_id]->get(bodies[j]);
                Vec3 tempLocalPos(pointOnBody[j], pointOnBody[numBodies+j], pointOnBody[2*numBodies+j]);
                Vec3 tempGlobalPos;
                Vec3 tempGlobalVel;
                osimModel[thread_id]->getSimbodyEngine().getPosition(*osimState[thread_id], tempRefParentBody, tempLocalPos, tempGlobalPos);
                osimModel[thread_id]->getSimbodyEngine().getVelocity(*osimState[thread_id], tempRefParentBody, tempLocalPos, tempGlobalVel);
                for (int k=0;k<3;k++)
                {
                    bodyPos[i + 3*numPts*j+numPts*k] = tempGlobalPos(k);
                    bodyVel[i + 3*numPts*j+numPts*k] = tempGlobalVel(k);
                }

				double tempGlobalAng[3][3];
				Vec3 tempGlobalAngVel;
				osimModel[thread_id]->getSimbodyEngine().getDirectionCosines(*osimState[thread_id], tempRefParentBody, tempGlobalAng);
				osimModel[thread_id]->getSimbodyEngine().getAngularVelocity(*osimState[thread_id], tempRefParentBody, tempGlobalAngVel);

				double theta_x = 0;
				double theta_y = -asin(tempGlobalAng[2][0]);
				double theta_z = 0;

				double cosTheta_y = cos(theta_y);
				if(cosTheta_y > 0.000001)
				{
					theta_x = atan2(tempGlobalAng[2][1],tempGlobalAng[2][2]);
					theta_z = atan2(tempGlobalAng[1][0],tempGlobalAng[0][0]);
				}
				else if(cosTheta_y < -0.000001)
				{
					theta_x = atan2(-tempGlobalAng[2][1],-tempGlobalAng[2][2]);
					theta_z = atan2(-tempGlobalAng[1][0],-tempGlobalAng[0][0]);
				}
				else
				{
					theta_z = 0;
					if(tempGlobalAng[2][0] > 0)
					{
						theta_x = atan2(tempGlobalAng[0][1],tempGlobalAng[0][2]);
					}
					else
					{
						theta_x = atan2(-tempGlobalAng[0][1],-tempGlobalAng[0][2]);
					}
				}

				bodyAng[i + 3*numPts*j] = theta_x;
				bodyAng[i + 3*numPts*j + numPts] = theta_y;
				bodyAng[i + 3*numPts*j + numPts*2] = theta_z;
				for (int kkk = 0; kkk < 3; kkk++)
				{
					bodyAngVel[i + 3*numPts*j + numPts*kkk] = tempGlobalAngVel(kkk);
				}
            }

            //delete x_row; // release memory for vectors of states for each time
			free(x_row); // release memory for vectors of states for each time
        }

       
    }
	else
	{
		mexErrMsgTxt("Error: Invalid number of input arguments\n");
	}
}