#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <OpenSim/OpenSim.h>
#include <InverseDynamicsSolver.h>
//#include <omp.h>
//#include <fstream> // For writing to files for debugging

using namespace OpenSim;
using namespace SimTK;
using namespace std;
//#define NUMTHREADS 14

/******************************************************************************/
/* Function and Static Variable Definitions ***********************************/

static Model *osimModel;
static State *osimState;
static InverseDynamicsSolver *idSolver;
static bool modelIsLoaded = false;

void ClearMemory(void)
{
	delete osimModel;
	delete idSolver;
	
	modelIsLoaded = false;
	mexPrintf("Cleared memory from opensimDeriv mex file.\n");
}

/*Main Function****************************************************************/
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	mexAtExit(ClearMemory); // exit function to clear memory allocated in heap

	ofstream debugFile;
	//debugFile.open("opensimID_OMPDebugFile.txt");
	//debugFile << "Starting Main.\n";
	//debugFile.close();
/*Load Model*******************************************************************/
	if (nrhs == 1) // Load model if only one right hand side argument
	{
		//debugFile.open("opensimID_OMPDebugFile.txt");
		//debugFile << "nrhs == 1.\n";
		//debugFile.close();
		// extract filenames
		string model_name = mxArrayToString(prhs[0]);
		if (modelIsLoaded == true)
		{
			ClearMemory();
		}

		osimModel = new Model(model_name); // name of Opensim model to load
		// Initialize state
		osimState = &osimModel->initSystem();
		modelIsLoaded = true;
		// Create an inverse dynamics solver
		idSolver = new InverseDynamicsSolver(*osimModel);
		
	}
	else if(nrhs == 5) // time, states, accels, controls, options vector
	{
		//debugFile.open("opensimID_OMPDebugFile.txt");
		//debugFile << "nrhs == 5.\n";
		//debugFile.close();
		if (modelIsLoaded == false)
		{
			mexErrMsgTxt("!!!No OpenSim model has been loaded!!!\n");
		}
		
/*Extracting Mex Function Inputs***********************************************/
		//debugFile.open("opensimID_OMPDebugFile.txt");
		//debugFile << "Extracting Mex Function Inputs.\n";
		//debugFile.close();
		const int numPts = mxGetM(prhs[0]); // get number of rows of time vector
		const int numStates = mxGetN(prhs[1]); // get number of states (columns in state matrix)
		const int numAccels = mxGetN(prhs[2]); // get number of accels
		const int numControls = mxGetN(prhs[3]); // get number of controls (columns in controls matrix)

		double *time = mxGetPr(prhs[0]); // time vector
		double *x = mxGetPr(prhs[1]); // states matrix
		double *accels = mxGetPr(prhs[2]); // accelerations matrix
		double *u = mxGetPr(prhs[3]); // controls matrix
		double *options = mxGetPr(prhs[4]); // options vector

/*Declaring Variables**********************************************************/
		// Declaring variables required for inverse dynamics solver
		//debugFile.open("opensimID_OMPDebugFile.txt");
		//debugFile << "Declaring variables for ID solver.\n";
		//debugFile.close();
		double *accels_row = (double*) mxCalloc(numAccels, sizeof(double));
		int numCoords = osimState->getNQ();
		plhs[0] = mxCreateDoubleMatrix(numPts,numCoords,mxREAL);
		double *idLoads = mxGetPr(plhs[0]);
		Vector idLoadsVec;
		Vector UDotVec;

		// Declaring variables required for marker positions
		//debugFile.open("opensimID_OMPDebugFile.txt");
		//debugFile << "Declaring vars for marker positions.\n";
		//debugFile.close();
		MarkerSet& refMarkerSet = osimModel->updMarkerSet();
		Array<std::string> tmpArrMarkerNames;
		refMarkerSet.getNames(tmpArrMarkerNames); // number of markers
		int numMarkers = tmpArrMarkerNames.getSize();
		int dims[3];
		dims[0] = numPts;
		dims[1] = numMarkers;
		dims[2] = 3; // three coordinates in space per marker
		plhs[1] = mxCreateNumericArray(3, dims, mxDOUBLE_CLASS, mxREAL);
		double* MarkerGlobalPos = mxGetPr(plhs[1]);

		// Declaring variables required for force reporting
		//debugFile.open("opensimID_OMPDebugFile.txt");
		//debugFile << "Declaring vars for force reporting.\n";
		//debugFile.close();
		const ForceSet& forces = osimModel->getForceSet(); // This does not contain gravity
        int nf = forces.getSize();
		Storage forceStore;

		double *x_row = (double*) mxCalloc(numStates, sizeof(double));
		double *u_row = (double*) mxCalloc(numControls, sizeof(double));

		// plhs[0] = mxCreateDoubleMatrix(numPts,numStates,mxREAL);
		// double *x_dot = mxGetPr(plhs[0]);

/*Parallel For Loop************************************************************/
		//debugFile.open("opensimID_OMPDebugFile.txt");
		//debugFile << "Start multithreading.\n";
		//debugFile.close();
		// start looping for all collocation points
		//#pragma omp parallel for num_threads(NUMTHREADS) // convert loop to parrallel using OpenMP
		for (int i=0; i<numPts; ++i)
		{
/*Set Up Loop Variables********************************************************/
			//debugFile.open("opensimID_OMPDebugFile.txt");
			//debugFile << "Set up loop vars.\n";
			//debugFile.close();

			// get a row vector from each of the matrices
			//debugFile.open("opensimID_OMPDebugFile.txt");
			//debugFile << "Create x and u row vectors.\n";
			//debugFile.close();
			for (int j=0;j<numStates;j++)
			{
				x_row[j] = x[i+numPts*j]; // irow+nrow*icol (column-wise indexing)
			}
			for (int j=0;j<numControls;j++)
			{
				u_row[j] = u[i+numPts*j]; // irow+nrow*icol (column-wise indexing)
			}

			// Set time
			osimState->setTime(time[i]);

			// Set states
			const Vector y(numStates, x_row);
			osimState->updY() = y;

			// Update state velocity calculations
			//debugFile.open("opensimID_OMPDebugFile.txt");
			//debugFile << "Update and define control vars.\n";
			//debugFile.close();
			osimModel->assemble(*osimState);

			// Define activation control vectors for each actuator
			const Vector newControls(numControls, u_row);

			// Get a writeable reference to model controls and set to new control values
			osimModel->updControls(*osimState) = newControls;
			
			// Done modifying model controls so mark as valid
			osimModel->markControlsAsValid(*osimState);

/*Get Inverse Dynamics Loads**************************************************/
			if (options[0] == 1) // if get residual loads set to true
			{
				//debugFile.open("opensimID_OMPDebugFile.txt");
				//debugFile << "Get ID Loads.\n";
				//debugFile.close();
				// get row vector of accelerations
				for (int jjj=0;jjj<numCoords;jjj++)
				{
					// accels_row[jjj] = accels[i*numCoords+jjj]; //(row-wise indexing)
					accels_row[jjj] = accels[i+numPts*jjj]; //(column-wise indexing)
				}

				// get Udot
				Vector UDotVec = Vector(numCoords, accels_row);

				// calling InverseDynamicsSolver
				//debugFile.open("opensimID_OMPDebugFile.txt");
				//debugFile << "Calling ID Solver.\n";
				//debugFile.close();
				idLoadsVec = idSolver->solve(*osimState,UDotVec); 

				// assign residual loads for each time to output matrix
				for (int j=0;j<numCoords;j++)
				{
					idLoads[i + numPts*j] = idLoadsVec[j]; 
				}
			}
/*Get Marker Positions*********************************************************/
			if (options[1] == 1) // if get marker global positions set to true
			{
				//debugFile.open("opensimID_OMPDebugFile.txt");
				//debugFile << "Get marker positions.\n";
				//debugFile.close();
			for(int j=0; j<numMarkers; j++)
				{
					Marker& tempRefMarker = refMarkerSet.get(tmpArrMarkerNames[j]);
					OpenSim::Body& tempRefMarkerParentBody = tempRefMarker.getBody();
					Vec3 tempMarkerLocalPos = tempRefMarker.getOffset();
					Vec3 tempMarkerGlobalPos;
					osimModel->getSimbodyEngine().getPosition(*osimState, tempRefMarkerParentBody, tempMarkerLocalPos, tempMarkerGlobalPos);
					for (int k=0;k<3;k++)
					{
						MarkerGlobalPos[i + numPts*j+numPts*numMarkers*k] = tempMarkerGlobalPos(k);
					}
				} 
			}
/*Get Force Values*************************************************************/
			if (options[2] == 1) // if get force values set to true
			{
				//debugFile.open("opensimID_OMPDebugFile.txt");
				//debugFile << "Get force values.\n";
				//debugFile.close();
				StateVector nextRow = StateVector(osimState->getTime());
				for(int j=0;j<nf;j++)
				{
					OpenSim::Force& nextForce = (OpenSim::Force&)forces[j];
					if (nextForce.isDisabled(*osimState)) continue;
					Array<double> values = nextForce.getRecordValues(*osimState);
					nextRow.getData().append(values);
				}
				forceStore.append(nextRow); 
			}
		}
		if (options[2] == 1) // if get force values set to true
		{
			//debugFile.open("opensimID_OMPDebugFile.txt");
			//debugFile << "Assign force values.\n";
			//debugFile.close();
			int numMeasures = forceStore.getSmallestNumberOfStates();
			plhs[2] = mxCreateDoubleMatrix(numPts,numMeasures,mxREAL);
			double *forceMeasures = mxGetPr(plhs[2]);
			double val;
			for (int i=0;i<numPts;i++)
			{
				for (int j=0;j<numMeasures;j++)
				{
					forceStore.getData(i,j,val);
					forceMeasures[i + numPts*j] = val;
				}
			}
		}
		mxFree(x_row); // release memory for vectors of states for each time
		mxFree(u_row);
		
	}
	else
	{
		mexErrMsgTxt("!!!Wrong inputs or outputs!!!\n");
	}

}


