#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <OpenSim/OpenSim.h>
#include <InverseDynamicsSolver.h>
#include <omp.h>

using namespace OpenSim;
using namespace SimTK;
using namespace std;
#define NUMTHREADS 14

/******************************************************************************/
/* Function and Static Variable Definitions ***********************************/

static Model *osimIDOMP_osimModel[NUMTHREADS];
static State *osimIDOMP_osimState[NUMTHREADS];
static InverseDynamicsSolver *osimIDOMP_idSolver[NUMTHREADS];
static bool osimIDOMP_modelIsLoaded = false;

void ClearMemory(void)
{
	for (int iii = 0; iii < NUMTHREADS; iii++)
	{
		delete osimIDOMP_osimModel[iii];
		//delete osimIDOMP_osimState[iii]; // NOTE (Nathan): Deleting the osimState variable causes Matalab to crash when clearing memory from within Matlab (i.e. clear all)
		delete osimIDOMP_idSolver[iii];
	}
	osimIDOMP_modelIsLoaded = false;
	mexPrintf("Cleared memory from opensimID_OMP mex file.\n");
}

/*Main Function****************************************************************/
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	
/*Load Model*******************************************************************/
	if (nrhs == 1) // Load model if only one right hand side argument
	{
		// extract filenames
		string model_name = mxArrayToString(prhs[0]);
		if (osimIDOMP_modelIsLoaded == true)
		{
			ClearMemory();
		}

		for (int iii = 0; iii < NUMTHREADS; iii++)
		{
			osimIDOMP_osimModel[iii] = new Model(model_name); // name of Opensim model to load
			// Initialize state
			osimIDOMP_osimState[iii] = &osimIDOMP_osimModel[iii]->initSystem();
			// Create an inverse dynamics solver
			osimIDOMP_idSolver[iii] = new InverseDynamicsSolver(*osimIDOMP_osimModel[iii]);
		}
		osimIDOMP_modelIsLoaded = true;
		mexAtExit(ClearMemory); // exit function to clear memory allocated in heap
		
	}
	else if(nrhs == 5) // time, states, accels, controls, options vector
	{
		if (osimIDOMP_modelIsLoaded == false)
		{
			mexErrMsgTxt("!!!No OpenSim model has been loaded!!!\n");
		}
		
/*Extracting Mex Function Inputs***********************************************/
		const int numPts = mxGetM(prhs[0]); // get number of rows of time vector
		const int numStates = mxGetN(prhs[1]); // get number of states (columns in state matrix)
		const int numAccels = mxGetN(prhs[2]); // get number of accels
		const int numControls = mxGetN(prhs[3]); // get number of controls (columns in controls matrix)

		double *time = mxGetPr(prhs[0]); // time vector
		double *x = mxGetPr(prhs[1]); // states matrix
		double *accels = mxGetPr(prhs[2]); // accelerations matrix
		double *u = mxGetPr(prhs[3]); // controls matrix
		double *options = mxGetPr(prhs[4]); // options vector

/*Declaring Variables**********************************************************/
		// Declaring variables required for inverse dynamics solver
		int numCoords = osimIDOMP_osimState[0]->getNQ();
		plhs[0] = mxCreateDoubleMatrix(numPts,numCoords,mxREAL);
		double *idLoads = mxGetPr(plhs[0]);

		// Declaring variables required for marker positions
		MarkerSet *refMarkerSet[NUMTHREADS];
		for(int iii = 0; iii < NUMTHREADS; iii++)
		{
			refMarkerSet[iii] = &osimIDOMP_osimModel[iii]->updMarkerSet();
		}

		//MarkerSet& refMarkerSet = osimIDOMP_osimModel->updMarkerSet();
		Array<std::string> tmpArrMarkerNames;
		refMarkerSet[0]->getNames(tmpArrMarkerNames); // number of markers
		int numMarkers = tmpArrMarkerNames.getSize();
		int dims[3];
		dims[0] = numPts;
		dims[1] = numMarkers;
		dims[2] = 3; // three coordinates in space per marker
		plhs[1] = mxCreateNumericArray(3, dims, mxDOUBLE_CLASS, mxREAL);
		double* MarkerGlobalPos = mxGetPr(plhs[1]);

/*Parallel For Loop************************************************************/
		#pragma omp parallel for num_threads(NUMTHREADS) // convert loop to parrallel using OpenMP
		// start looping for all collocation points
		for (int i=0; i<numPts; ++i)
		{
/*Set Up Loop Variables********************************************************/
			
			int thread_id = omp_get_thread_num(); // Get the thread identification

			// get a row vector from each of the matrices
			// NOTE (Nathan): Don't use mxCalloc as it doesn't play well with omp
			// These are declared inside of the for loop so that they are treated as private vars by omp
			double *x_row = (double*) calloc(numStates, sizeof(double));
			double *u_row = (double*) calloc(numControls, sizeof(double));
			
			for (int j=0;j<numStates;j++)
			{
				x_row[j] = x[i+numPts*j]; // irow+nrow*icol (column-wise indexing)
			}
			for (int j=0;j<numControls;j++)
			{
				u_row[j] = u[i+numPts*j]; // irow+nrow*icol (column-wise indexing)
			}

			// Set time
			osimIDOMP_osimState[thread_id]->setTime(time[i]);

			// Set states
			const Vector yStateVector(numStates, x_row);
			//Vector yStateVector(numStates, x_row);
			osimIDOMP_osimState[thread_id]->updY() = yStateVector;

			// Update state velocity calculations
			osimIDOMP_osimModel[thread_id]->assemble(*osimIDOMP_osimState[thread_id]);
			
			//osimIDOMP_osimModel->assemble(*osimIDOMP_osimState);

			// Define activation control vectors for each actuator
			const Vector newControls(numControls, u_row);
			//Vector newControls(numControls, u_row);
			// Get a writeable reference to model controls and set to new control values
			osimIDOMP_osimModel[thread_id]->updControls(*osimIDOMP_osimState[thread_id]) = newControls;
			
			// Done modifying model controls so mark as valid
			osimIDOMP_osimModel[thread_id]->markControlsAsValid(*osimIDOMP_osimState[thread_id]);
			
/*Get Inverse Dynamics Loads**************************************************/
			if (options[0] == 1) // if get residual loads set to true
			{
				// get row vector of accelerations
				double *accels_row = (double*) calloc(numAccels, sizeof(double));
				for (int jjj=0;jjj<numCoords;jjj++)
				{
					// accels_row[jjj] = accels[i*numCoords+jjj]; //(row-wise indexing)
					accels_row[jjj] = accels[i+numPts*jjj]; //(column-wise indexing)
				}

				// get Udot
				Vector UDotVec = Vector(numCoords, accels_row);

				// calling InverseDynamicsSolver
				Vector idLoadsVec;
				idLoadsVec = osimIDOMP_idSolver[thread_id]->solve(*osimIDOMP_osimState[thread_id],UDotVec);

				// assign residual loads for each time to output matrix
				for (int j=0;j<numCoords;j++)
				{
					idLoads[i + numPts*j] = idLoadsVec[j];
				}
				free(accels_row);
			}
/*Get Marker Positions*********************************************************/
			if (options[1] == 1) // if get marker global positions set to true
			{
			for(int j=0; j<numMarkers; j++)
				{
					Marker& tempRefMarker = refMarkerSet[thread_id]->get(tmpArrMarkerNames[j]);
					OpenSim::Body& tempRefMarkerParentBody = tempRefMarker.getBody();
					Vec3 tempMarkerLocalPos = tempRefMarker.getOffset();
					Vec3 tempMarkerGlobalPos;
					osimIDOMP_osimModel[thread_id]->getSimbodyEngine().getPosition(*osimIDOMP_osimState[thread_id], tempRefMarkerParentBody, tempMarkerLocalPos, tempMarkerGlobalPos);
					for (int k=0;k<3;k++)
					{
						MarkerGlobalPos[i + numPts*j+numPts*numMarkers*k] = tempMarkerGlobalPos(k);
					}
				} 
			}
			free(x_row); // release memory for vectors of states for each time
			free(u_row);
			
		}
	}
	else
	{
		mexErrMsgTxt("!!!Wrong inputs or outputs!!!\n");
	}

}


