#include "mex.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <OpenSim/OpenSim.h>
#include <omp.h>

using namespace OpenSim;
using namespace SimTK;
using namespace std;
#define NTHREADS 14

//______________________________________________________________________________


static Model *osimModel[NTHREADS];
static State *osimState[NTHREADS];
static bool modelIsLoaded = false;

void ClearMemory(void)
{
    for (int i=0; i<NTHREADS; ++i)
                delete osimModel[i];
    modelIsLoaded = false;
    mexPrintf("Cleared memory from opensimPointKin mex file.\n");
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	mexAtExit(ClearMemory); // exit function to clear memory allocated in heap
    
    if (nrhs == 1) // Load model if only one right hand side argument
    {
        if (modelIsLoaded == true)
            ClearMemory();
        // extract filenames
        string model_name = mxArrayToString(prhs[0]);
        for (int i=0; i<NTHREADS; ++i)
        {
            osimModel[i] = new Model(model_name); // name of Opensim model to load
            // Initialize state
            osimState[i] = &osimModel[i]->initSystem();
        }
        modelIsLoaded = true;
    }
    else if(nrhs == 4) // time, states, spring locations, spring bodies by number, body index for spring location
    {
        if (modelIsLoaded == false)
        {
            mexErrMsgTxt("!!!No OpenSim model has been loaded!!!\n");
        }
        
        const int numPts = mxGetM(prhs[0]); // get number of rows of time vector
        const int numStates = mxGetN(prhs[1]); // get number of states (columns in state matrix)
        const int numSprings = mxGetM(prhs[3]); // get number of bodies springs are located on
        
        double *time = mxGetPr(prhs[0]); // time vector
        double *x = mxGetPr(prhs[1]); // states matrix
        double *sp = mxGetPr(prhs[2]); // spring locations within body
        double *bodies = mxGetPr(prhs[3]); // body number index for springs
        
        OpenSim::BodySet *refBodySet[NTHREADS];
        // define variables used in determining spring locations and velocities
        for (int i=0; i<NTHREADS; ++i)
        {
            refBodySet[i] = &osimModel[i]->updBodySet();
        }
        
        // create output matrices
        plhs[0] = mxCreateDoubleMatrix(numPts,numSprings*3,mxREAL);
        plhs[1] = mxCreateDoubleMatrix(numPts,numSprings*3,mxREAL);
        double *sp_pos = mxGetPr(plhs[0]);
        double *sp_vel = mxGetPr(plhs[1]);
        
        // start looping for all collocation points
        #pragma omp parallel for num_threads(NTHREADS)
        for (int i=0; i<numPts; ++i)
        {
            int thread_id = omp_get_thread_num();

            // change based on number of threads being used
            //double *x_row = new double [numStates];
            double *x_row = (double*) calloc(numStates, sizeof(double));

//             Model *osimModelThread = osimModel[thread_id];
//             State *osimStateThread = osimState[thread_id];
            
            // get a row vector from each of the matrices
            for (int j=0;j<numStates;j++)
            {
                x_row[j] = x[i+numPts*j]; // irow+nrow*icol (column-wise indexing)
            }

            // Set time
            osimState[thread_id]->setTime(time[i]);

            // Set states
            const Vector y(numStates, x_row);
            osimState[thread_id]->updY() = y;

            // Update state velocity calculations
//             osimModel->assemble(*osimState);
            osimModel[thread_id]->calcMassCenterVelocity(*osimState[thread_id]);
            
            // get spring positions and velocities
            for(int j=0; j<numSprings; j++)
            {
                OpenSim::Body& tempRefParentBody = refBodySet[thread_id]->get(bodies[j]);
                Vec3 tempLocalPos(sp[j*3], sp[j*3+1], sp[j*3+2]);
                Vec3 tempGlobalPos;
                Vec3 tempGlobalVel;
                osimModel[thread_id]->getSimbodyEngine().getPosition(*osimState[thread_id], tempRefParentBody, tempLocalPos, tempGlobalPos);
                osimModel[thread_id]->getSimbodyEngine().getVelocity(*osimState[thread_id], tempRefParentBody, tempLocalPos, tempGlobalVel);
                for (int k=0;k<3;k++)
                {
                    sp_pos[i + 3*numPts*j+numPts*k] = tempGlobalPos(k);
                    sp_vel[i + 3*numPts*j+numPts*k] = tempGlobalVel(k);
                }
            }
            //delete x_row; // release memory for vectors of states for each time
			free(x_row); // release memory for vectors of states for each time
        }

       
    }
}


